class Vector
  class DimensionMismatchError < ArgumentError; end

  attr_reader :dimension

  def initialize(data)
    @dimension = data.size
    @data = data.dup
  end

  def +(other)
    operator(:+, other) { |new_data| Vector.new new_data }
  end

  def -(other)
    operator(:-, other) { |new_data| Vector.new new_data }
  end

  def *(other)
    operator(:*, other) { |new_data| new_data.inject(:+) }
  end

  def each(&blk)
    @data.each(&blk)
  end

  def map(&blk)
    Vector.new @data.map(&blk)
  end

  def to_a
    @data.dup
  end

  def to_s
    "(#{@data.join(", ")})"
  end

  def ==(other)
    @data == other.data
  end

  protected
  def data
    @data
  end

  private
  def operator(name, other)
    if other.kind_of? Vector

      unless other.dimension == self.dimension
        msg = "Dimension mismatch, got #{other.dimension}, expected #{self.dimension}"
        raise DimensionMismatchError, msg
      end

      other_data = other.data

      new_data = []
      other_data.each_with_index do |e, i|
        new_data[i] = @data[i].send(name, e)
      end

      yield new_data

    elsif other.kind_of? Numeric
      map { |e| other.send(name, e) }
    else
      raise TypeError, "Don't know how to do #{name} with #{other.class}."
    end
  end
end

a = Vector.new [1, 2, 3]
b = Vector.new [4, 5, 6]
c = Vector.new [1, 2, 3]

puts "a: #{a}"
puts "b: #{b}"
puts "c: #{c}"

puts "a == b: #{a == b}"
puts "a == c: #{a == c}"

puts "a + b: #{a + b}"
puts "a - b: #{a - b}"
puts "a * b: #{a * b}"

puts "a + 3: #{a + 3}"
puts "b + 3: #{b + 3}"

puts "a.map { |e| e * e}: #{a.map { |e| e * e}}"
